use config::Config;
use xcb;
use ui::ext;
use ui::ext::ConnectionExt;
use ui::font;
use ui::tray;
use ui::x11;

use std::cell::RefCell;
use std::rc::Rc;

pub struct Panel {
    pub conn: x11::Connection,
    pub window: xcb::Window,
    pub width: u16,
    tray: RefCell<Option<tray::Tray>>,
    fonts: font::FontLoader,
    picture: xcb::render::Picture,
    finishing: RefCell<bool>
}

impl Panel {
    pub fn new(conn: x11::Connection, cfg: &Config) -> Rc<Panel> {
        let val = cfg.lookup("bar.fonts").unwrap();
        let fonts = val.as_slice().unwrap();
        let fonts = fonts.iter().flat_map(|elem| elem.as_str());
        let mut font_loader = font::FontLoader::new(conn.clone_connection());
        for font in fonts {
            font_loader.load(font);
        }

        let window = conn.generate_id();
        let picture = conn.generate_id();

        let panel = Rc::new(Panel {
            window: window,
            width: conn.default_screen().width,
            fonts: font_loader,
            conn: conn,
            tray: RefCell::new(None),
            picture: picture,
            finishing: RefCell::new(false)
        });
        let tray = tray::Tray::new(Rc::downgrade(&panel));
        *panel.tray.borrow_mut() = Some(tray);
        panel
    }

    pub fn create(&self) {
        let conn = &self.conn;
        let screen = conn.default_screen();

        xcb::create_window(
            conn,
            xcb::COPY_FROM_PARENT as u8,
            self.window,
            screen.root,
            0, 0,
            screen.width, 20,
            0,
            xcb::WINDOW_CLASS_INPUT_OUTPUT as u16,
            xcb::COPY_FROM_PARENT,
            &[
                (xcb::CW_BACK_PIXEL, 0xFF000000),
                (xcb::CW_EVENT_MASK, xcb::EVENT_MASK_EXPOSURE | xcb::EVENT_MASK_PROPERTY_CHANGE | xcb::EVENT_MASK_STRUCTURE_NOTIFY)
            ]
        );
        self.set_property(
            conn.atom(x11::_NET_WM_WINDOW_TYPE),
            xcb::ATOM_ATOM,
            32,
            &[conn.atom(x11::_NET_WM_WINDOW_TYPE_DOCK)]
        );
        xcb::map_window(conn, self.window);
        xcb::render::create_picture(&conn, self.picture, self.window, conn.get_pict_format(ext::PictFormat::RGB24), &[
            (xcb::render::CP_POLY_EDGE, xcb::render::POLY_EDGE_SMOOTH),
            (xcb::render::CP_POLY_MODE, xcb::render::POLY_MODE_IMPRECISE)
        ]);
        conn.flush();
    }

    pub fn set_property<T>(&self, name: xcb::Atom, type_: xcb::Atom, format: u8, data: &[T]) {
        xcb::change_property(
            &self.conn,
            xcb::PROP_MODE_REPLACE as u8,
            self.window,
            name,
            type_,
            format,
            data
        );
    }

    pub fn handle_event(&self, event: xcb::GenericEvent) -> bool {
        let tray = self.tray.borrow_mut().as_mut().unwrap().handle_event(&event);
        if event.response_type() == xcb::EXPOSE {
            self.draw_text("Hello 世界!");
        }
        if self.is_finishing() && event.response_type() == xcb::DESTROY_NOTIFY {
            let event: &xcb::DestroyNotifyEvent = xcb::cast_event(&event);
            if event.window() == self.window {
                return true
            }
        }
        tray
    }

    pub fn create_pen(&self, r: u16, g: u16, b: u16, a: u16) -> xcb::render::Picture {
        let color = xcb::render::Color::new(r, g, b, a);
        let format = self.conn.get_pict_format(ext::PictFormat::ARGB32);
        let pixmap = self.conn.generate_id();
        xcb::create_pixmap(&self.conn, 32, pixmap, self.window, 1, 1);
        let picture = self.conn.generate_id();
        xcb::render::create_picture(
            &self.conn,
            picture,
            pixmap,
            format,
            &[(xcb::render::CP_REPEAT, xcb::render::REPEAT_NORMAL)]
        );
        xcb::render::fill_rectangles(
            &self.conn,
            xcb::render::PICT_OP_OVER as u8,
            picture,
            color,
            &[xcb::Rectangle::new(0, 0, 1, 1)]
        );
        xcb::free_pixmap(&self.conn, pixmap);
        self.conn.flush();
        picture
    }

    pub fn draw_text(&self, name: &str) {
        let pen = self.create_pen(0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF);
        let text = self.fonts.create_renderable_text(name);
        let baseline = 20 - self.fonts.default_offset(20);
        text.render(&self.conn, pen, self.picture, 0, baseline as u16);
        self.conn.flush();
    }

    pub fn is_finishing(&self) -> bool {
        *self.finishing.borrow()
    }

    pub fn finish(&self) {
        *self.finishing.borrow_mut() = true;
        self.tray.borrow_mut().as_mut().unwrap().finish();
        xcb::destroy_window(&self.conn, self.window);
        self.conn.flush();
    }
}
